/* 
 * Morpheus by Kuba "BlackWine" Winnicki
 * 
 * 3d modeller object viewer for GNOME <http://www.gnome.org>
 * and Mesa <http://www.mesa3d.org>/OpenGL[TM] <http://www.opengl.org>
 *
 * OpenGL[TM] is a trademark of SGI.
 *
 * (C) 1999 Kuba Winnicki <mailto:bw@sexcity.pl>
 *
 * Under terms of GNU GPL
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <morph/morph.h>

#undef	MAX
#define	MAX(a, b)  (((a) > (b)) ? (a) : (b))

#undef	MIN
#define	MIN(a, b)  (((a) < (b)) ? (a) : (b))

MorphModel*	model = NULL;

int		prevx, prevy, prevz, prev_stick_z;

GLint		bilerp = GL_NEAREST; /* GL_LINEAR gave strange effects
					on texture wrap borders */

struct Camera {
	GLsizei		ViewportW, ViewportH;
	GLfloat		Head, Pitch, Bank,
			Dist, Scale,
			CenterX, CenterY, CenterZ;
} gpCam = { 512, 384 };

/*
GLfloat l0_pos[] = { 0.0, 1000.0, 0.0, 1.0 };
GLfloat l1_dns[] = { 1.0, 1.0, 1.0, 1.0 };
*/

GLboolean	lmb, mmb, rmb,
		textures = GL_FALSE;

void
gpGLInit( void )
{
	glCullFace( GL_BACK );
	glEnable( GL_CULL_FACE );
/*	glBlendFunc( GL_SRC_ALPHA_SATURATE, GL_ONE );
*/
	glEnable( GL_DEPTH_TEST );

	glDepthFunc( GL_LESS );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glScalef( gpCam.Scale, gpCam.Scale, gpCam.Scale );
}

void
glarea_set_bg( GLushort* bg )
{
	glClearColor( (GLfloat )bg[ 0 ] / 65535,
		      (GLfloat )bg[ 1 ] / 65535,
		      (GLfloat )bg[ 2 ] / 65535, 
		      (GLfloat )bg[ 3 ] / 65535 );
}

/* Find scale and center for display normalization */

void
FindCenterScale( GLfloat* _vertices, int _vertices_len )
{
	int	i = _vertices_len;
	GLfloat	max_x, min_x, max_y, min_y, max_z, min_z;

	i >>= 2;

	_vertices += i;
	max_z = min_z = *--_vertices;
	max_y = min_y = *--_vertices;
	max_x = min_x = *--_vertices;

	for( i -= 4; i > 0; i -= 3 ) {	/* all scanned? */
		--_vertices;
		max_z = MAX( max_z, *_vertices );
		min_z = MIN( min_z, *_vertices );
		--_vertices;
		max_y = MAX( max_y, *_vertices );
		min_y = MIN( min_y, *_vertices );
		--_vertices;
		max_x = MAX( max_x, *_vertices );
		min_x = MIN( min_x, *_vertices );
	}
/*
	printf( "mini max:\n%f %f\n%f %f\n%f %f\n",
			min_x, max_x, min_y, max_y, min_z, max_z );
*/
	gpCam.CenterX	= min_x + ( max_x - min_x ) / 2;
	gpCam.CenterY	= min_y + ( max_y - min_y ) / 2;
	gpCam.CenterZ	= min_z + ( max_z - min_z ) / 2;
	gpCam.Scale	= 1500.0 / MAX( MAX( gpCam.CenterX - min_x,
					     gpCam.CenterY - min_y ),
					gpCam.CenterZ - min_z );
/*
	printf( "center/scale:\n(%f %f %f) * %f\n",
		gpCam.CenterX, gpCam.CenterY, gpCam.CenterZ, gpCam.Scale);
*/
}

int
gpLoadScene( const unsigned char* file_name )
{
	if( model )
		morph_model_unref( model );
	
	if( file_name )
		model = morph_model_load( file_name );

	if( !model ) {
		fprintf( stderr, "Unable to load %s\n", file_name );
		fflush( stderr );
		return 1;
	}

	/* Set Camera up */
	FindCenterScale( model->vertices, model->vertices_len );
	gpCam.Head	= -30;
	gpCam.Pitch	= -20;
	gpCam.Dist	= -6000;

	if( model->triangle_indices )
		morph_calc_vnormals( model );

/*	for( i = 1; i < model->surface_count; i++ ) {
		printf( "Surface %2d: %s\n", i, model->surfaces[ i ].name );
	}
*/
	return 0;
}

/* Load textures */

/*
int 
meusLoadTextures()
{
	FILE*		fd_tex;
	const char*	fd_name = "../pixmaps/suckmychocolatesaltyballs.rgb";
*/
	/* Bind textures and set their capabilities */
/*	
	glBindTexture( GL_TEXTURE_2D, 1 );
	glTexImage2D( GL_TEXTURE_2D, 0,
		      GL_INTENSITY8,
		      64, 64, 0,
		      GL_RGB,
		      GL_UNSIGNED_BYTE,
		      wood_tex );
	
	glTexParameteri( GL_TEXTURE_2D,
			 GL_TEXTURE_MIN_FILTER, bilerp );
	glTexParameteri( GL_TEXTURE_2D,
			 GL_TEXTURE_MAG_FILTER, bilerp );
	glTexParameteri( GL_TEXTURE_2D,
			 GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D,
			 GL_TEXTURE_WRAP_T, GL_REPEAT );

}

*/

/* Here we render an object */

void
gpDrawScene() /* draw_view( GdkWidget* view ) */
{
	int		i, j;
	GLushort*	_indices;

	if( !model )
		return;
/*	glClear( GL_COLOR_BUFFER_BIT );

	glDisable( GL_CULL_FACE );
	glFrontFace( GL_CW );

	glLightfv( GL_LIGHT0, GL_POSITION, l0_pos );
	glLightModelf( GL_LIGHT_MODEL_LOCAL_VIEWER, 1.0 );
*/
	glShadeModel( GL_SMOOTH );
	glEnable( GL_LIGHT0 );
	glEnable( GL_LIGHTING );
	glEnable( GL_NORMALIZE );
	glEnable( GL_DITHER );
/*
	if( textures )
		glEnable( GL_TEXTURE_2D );
*/

#ifdef GL_VERSION_1_1
	glVertexPointer( 3, GL_FLOAT, 0, model->vertices );
	glNormalPointer( GL_FLOAT, 0, model->vnormals );

	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_NORMAL_ARRAY );
#endif

	_indices = model->triangle_indices;

	for( j = 0; j < model->surface_count; j++ ) {
		if( !model->surfaces[ j ].triangle_count )
			continue;

		if( model->surfaces[ j ].flags & MORPH_DBL_SIDE )
			glDisable( GL_CULL_FACE );
		else
			glEnable( GL_CULL_FACE );

		glMaterialiv( GL_FRONT_AND_BACK, GL_SPECULAR,
			      model->surfaces[ j ]._specular );
		glMaterialiv( GL_FRONT_AND_BACK, GL_DIFFUSE,
			      model->surfaces[ j ]._diffuse );

		glMaterialiv( GL_FRONT, GL_SHININESS,
			      &model->surfaces[ j ].specular_exp );
/*
		if( textures ) {
			glBindTexture( GL_TEXTURE_2D, 17 );
			glEnableClientState( GL_TEXTURE_COORD_ARRAY );
			glTexCoordPointer( 2, GL_FLOAT, 0, model->tex_coords );
		}
*/
#ifndef GL_VERSION_1_1
		glDrawElements( GL_TRIANGLES,
				model->surfaces[ j ].triangle_count * 3,
				GL_UNSIGNED_SHORT,
				_indices );
		_indices += model->surfaces[ j ].triangle_count * 3;
#else
		glBegin( GL_TRIANGLES );
		for( i = model->surfaces[ j ].triangle_count * 3; i > 0; i-- ) {
			glNormal3f(
				model->vnormals[ *_indices * 3 ],
				model->vnormals[ *_indices * 3 + 1 ],
				model->vnormals[ *_indices * 3 + 2 ] );
			glVertex3f(
				model->vertices[ *_indices * 3 ],
				model->vertices[ *_indices * 3 + 1 ],
				model->vertices[ *_indices * 3 + 2 ] );
			_indices++;
		}
		glEnd();
#endif
	}
/*
	if( textures )
		glDisable( GL_TEXTURE_2D );
*/
#ifdef GL_VERSION_1_1
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );
	glDisableClientState( GL_VERTEX_ARRAY );
#endif
	glDisable( GL_LIGHT0 );
	glDisable( GL_LIGHTING );
	glDisable( GL_NORMALIZE );
}

void
gpDisplay( void )
{
/*	glEnable( GL_BLEND );
	glEnable( GL_POLYGON_SMOOTH );
*/	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glEnable( GL_DEPTH_TEST );

	glLoadIdentity();
	glTranslatef( 0.0, 0.0, gpCam.Dist );

	glRotatef( gpCam.Head,  0.0, 1.0, 0.0 );
	glRotatef( gpCam.Pitch, 1.0, 0.0, 0.0 );
	glRotatef( gpCam.Bank,  0.0, 0.0, 1.0 );
	glScalef( gpCam.Scale, gpCam.Scale, gpCam.Scale );
	glTranslatef( -gpCam.CenterX, -gpCam.CenterY, -gpCam.CenterZ );

	gpDrawScene();
}

void
reshape( int w, int h )
{
	glViewport( 0, 0, (GLsizei )w, (GLsizei )h );

	gpCam.ViewportW = w;
	gpCam.ViewportH = h;

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 30.0, (GLfloat )w / (GLfloat )h, 64.0, 16384.0 );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glScalef( gpCam.Scale, gpCam.Scale, gpCam.Scale );

	glEnable( GL_LIGHT0 );
/*	glLightfv( GL_LIGHT0, GL_POSITION, l0_pos );
	glLightfv( GL_LIGHT1, GL_DIFFUSE, l1_dns );
	glLightfv( GL_LIGHT1, GL_SPECULAR, l1_dns );
*/
}

#define MIN_Y 0
#define MAX_Y 360
#define MIN_Z -8000
#define MAX_Z -400

void
motion( int x, int y )
{
	int d;

	if( lmb ) {
		gpCam.Head -= prevx - x;
		if( gpCam.Head < 0 )
			gpCam.Head += 360;
		if( gpCam.Head > 360 )
			gpCam.Head -= 360;

		d = prevy - y;
		gpCam.Pitch -= d * cos( gpCam.Head * ( M_PI / 180 ) );
		if( gpCam.Pitch < 0 )
			gpCam.Pitch += 360;
		if( gpCam.Pitch > 360 )
			gpCam.Pitch -= 360;

		gpCam.Bank -= d * sin( gpCam.Head * ( M_PI / 180 ) );
		if( gpCam.Bank < 0 )
			gpCam.Bank += 360;
		if( gpCam.Bank > 360 )
			gpCam.Bank -= 360;
	}

	if( mmb ) {
		gpCam.Dist += ( prevy - y ) * 4;

		if( gpCam.Dist < MIN_Z )
			gpCam.Dist = MIN_Z;
		if( gpCam.Dist > MAX_Z )
			gpCam.Dist = MAX_Z;
	}

	prevx = x;
	prevy = y;
	prevz = y;
}

void
keyboard( unsigned char key, int x, int y )
{
	switch( key ) {

		/* Generic stuff */

		case 27:
			exit( 0 );
			break;
		default:
			break;
	}
//	printf( "%d %d %d %c\n", key, x, y, key );
}

void
sweep( void )
{
	morph_model_unref( model );
}
