/*
 * Morpheus 0.1.0
 *
 * (C) 1999 Kuba "BlackWine" Winnicki <mailto:blackwine@optimus.wroc.pl>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtkgl/gtkglarea.h>
#include "glengine.h"
/*
#include <stdio.h>
#include <stdlib.h>
*/
#include "interface.h"
#include "support.h"
#include "main.h"

enum targets_enum {
	TARGET_URIS,
	TARGET_COLOR
};

gint
init( GtkWidget *widget )
{
	const guint16	bg[] = { 0x6666, 0x8000, 0x9999, 0xffff };

	if( gtk_gl_area_make_current( GTK_GL_AREA( widget ) ) ) {
		gpGLInit();
		glarea_set_bg( bg );
		reshape( widget->allocation.width, widget->allocation.height );
	}
	return TRUE;
}


gint
draw( GtkWidget *widget, GdkEventExpose *event )
{
      	/* Draw only last expose. */
	if( event->count > 0 )
		return TRUE;
	
	if( gtk_gl_area_make_current( GTK_GL_AREA( widget ) ) ) {
		reshape( widget->allocation.width, widget->allocation.height );
		gpDisplay();
		gtk_gl_area_swapbuffers( GTK_GL_AREA( widget ) );

	}

      	return TRUE;
}

gint
_reshape( GtkWidget *widget, GdkEventConfigure *event )
{
	if( gtk_gl_area_make_current( GTK_GL_AREA( widget ) ) )
		glViewport( 0, 0,
			    widget->allocation.width,
			    widget->allocation.height );
	return TRUE;
}

gint
button( GtkWidget* widget, GdkEventButton* event, GtkWidget* menu )
{
	if( event->button != 3 ) {
		prevx = event->x;
		prevy = event->y;
		return TRUE;
	} else {
		gtk_menu_popup( GTK_MENU( menu ), NULL, NULL,
				NULL, NULL, event->button, time( NULL ) );
	}

	return FALSE;
}

gint
_motion( GtkWidget* widget, GdkEventMotion* event )
{
	int		x, y;
	GdkModifierType	state;

      	if( event->is_hint )
		gdk_window_get_pointer( event->window, &x, &y, &state );
	else {
		x	= event->x;
		y	= event->y;
		state	= event->state;
	}

	if( state & GDK_BUTTON1_MASK ) {
		lmb = TRUE;
		motion( x, y );
		lmb = FALSE;

		gtk_widget_draw( widget, NULL );
	}

	if( state & GDK_BUTTON2_MASK ) {
		mmb = TRUE;
		motion( x, y );
		mmb = FALSE;

		gtk_widget_draw( widget, NULL );
	}

	prevx = x;
	prevy = y;
	prevz = y;

	return TRUE;
}

static void  
fed_with_drag_data( GtkWidget*		widget,
		    GdkDragContext*	context, 
		    gint		x,
		    gint		y,
		    GtkSelectionData*	selection_data,
		    guint		info,
		    guint		time )
{

	switch( info ) {
		case TARGET_URIS: {
			GList*	uris,
			     *	copy;

			uris = gnome_uri_list_extract_filenames(
					(char *)selection_data->data );

			/* FIXME: we need a mdi or something to really
			 * support uri list,
			 * FIXME: set title here set title there, make a single wrapper for all */

			for( copy = uris; copy; copy = copy->next ) {
				gpLoadScene( copy->data );
				gtk_window_set_title( GTK_WINDOW( morpheus.app ),
						      g_basename( copy->data ) );
			}

			gnome_uri_list_free_strings( uris );
			gtk_widget_draw( widget, NULL );
			break;
		}
		case TARGET_COLOR: {
			guint16*	data = (guint16 *)selection_data->data;
			int		i;

			if( selection_data->length != 8 )
				return;

			glarea_set_bg( data );
			gtk_widget_draw( widget, NULL );
			break;
		}
	}
}

int
main( int argc, char *argv[] )
{
	GtkWidget*	app1,
		 *	dock1,
		 *	glarea,
		 *	glarea_menu;

      	int	attrlist[] = {
			GDK_GL_RGBA,
			GDK_GL_DOUBLEBUFFER,
			GDK_GL_DEPTH_SIZE, 1,
			GDK_GL_NONE
		};

	static GtkTargetEntry target_table[] = {
		{ "text/uri-list",		0, TARGET_URIS },
		{ "application/x-color",	0, TARGET_COLOR }
	};

	if( gdk_gl_query() == FALSE ) {
		g_print( "no OpenGL capability\n" );
		return 0;
	}
	
/*	if( gpLoadScene( argv[ 1 ] ) )
		return 0;
*/
	gpLoadScene( argv[ 1 ] );
	bindtextdomain( PACKAGE, PACKAGE_LOCALE_DIR );
	textdomain( PACKAGE );

	gnome_init( "morpheus", VERSION, argc, argv );

	app1 = create_app1();
	morpheus.app = app1;

	glarea_menu = create_glarea_menu1();

	if( argv[ 1 ] )
		gtk_window_set_title( GTK_WINDOW( morpheus.app ),
				      g_basename( argv[ 1 ] ) );

	gtk_window_set_default_size( GTK_WINDOW( app1 ), 400, 400 );

	gtk_signal_connect( GTK_OBJECT( app1 ), "delete_event",
			    GTK_SIGNAL_FUNC( gtk_main_quit ), NULL );

	glarea = GTK_WIDGET( gtk_gl_area_new( attrlist ) );

	morpheus.glarea = glarea;

	/* Events for widget must be set before X Window is created */
	gtk_widget_set_events(	GTK_WIDGET( glarea ),
				  GDK_EXPOSURE_MASK
				| GDK_BUTTON_PRESS_MASK
				| GDK_BUTTON_RELEASE_MASK
				| GDK_POINTER_MOTION_MASK
				| GDK_POINTER_MOTION_HINT_MASK );

	gtk_signal_connect( GTK_OBJECT( glarea ), "expose_event",
			    GTK_SIGNAL_FUNC( draw ), NULL );
	gtk_signal_connect( GTK_OBJECT( glarea ), "configure_event",
			    GTK_SIGNAL_FUNC( _reshape ), NULL );
	gtk_signal_connect( GTK_OBJECT( glarea ), "realize",
			    GTK_SIGNAL_FUNC( init ), NULL );
	gtk_signal_connect( GTK_OBJECT( glarea ), "motion_notify_event",
	  		    GTK_SIGNAL_FUNC( _motion ), NULL );
	gtk_signal_connect( GTK_OBJECT( glarea ), "button_press_event",
		  	    GTK_SIGNAL_FUNC( button ), glarea_menu );
	gtk_signal_connect( GTK_OBJECT( glarea ), "drag_data_received",
			    GTK_SIGNAL_FUNC( fed_with_drag_data ), NULL );

	gtk_drag_dest_set( GTK_WIDGET( glarea ),
			   GTK_DEST_DEFAULT_MOTION |
			   GTK_DEST_DEFAULT_HIGHLIGHT |
			   GTK_DEST_DEFAULT_DROP,
			   target_table, 3,
			   GDK_ACTION_COPY );

//	gtk_widget_set_usize( GTK_WIDGET( glarea ), 320, 240 );

	dock1 = GNOME_APP( app1 )->dock;

/*	gtk_container_add( GTK_CONTAINER( dock1 ), GTK_WIDGET( glarea ) );//FIXME!
*/
	gtk_widget_ref( glarea );
	gtk_object_set_data_full( GTK_OBJECT( app1 ), "glarea", glarea,
                            	  ( GtkDestroyNotify )gtk_widget_unref );
	gnome_app_set_contents( GNOME_APP( app1 ), glarea );

//	gtk_widget_hide( GTK_WIDGET( list->next ) );
	gtk_widget_show( GTK_WIDGET( glarea ) );
	gtk_widget_show( app1 );

	gtk_main();

	sweep();
	return 0;
}

